"""
### geometric_correction.py ###

Functions for correction of image distortion by estimating an affime transformation from an FCC reference lattice (template)
"""
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from sklearn.neighbors import KDTree
import random
from skimage import transform

import grain_boundary_reconstruction.tools as tools

#####

def create_bulk_templates(ideal_bulk_df, projection_angles_deg, side='left'):
    '''returns list of dataframes with templates created from projections of the ideal bulk structure.
    GB plane set to 0 on z-axis.'''
    template_list=[]    
    for proj_angle in projection_angles_deg:
        ideal_bulk_df_proj = tools.find_projected_coordinates(ideal_bulk_df, proj_angle)
        max_dist=0.1
        ideal_bulk_df_proj_clustered = tools.cluster_2D_projection(ideal_bulk_df_proj, max_dist)
        
        ##Shift GB to 0
        if side=='left':
            ideal_bulk_df_proj_clustered.z *= -1
            ideal_bulk_df_proj_clustered.z -= np.max(ideal_bulk_df_proj_clustered.z)
        if side=='right':
            ideal_bulk_df_proj_clustered.z -= np.min(ideal_bulk_df_proj_clustered.z)
            
        template_list.append(ideal_bulk_df_proj_clustered)
    return template_list

def find_layer(refined_peaks_df, side, dz=5):
    '''returns df with layer atoms, at least 3 stk''' 
    if side=='left':
        z_layer_image_df = refined_peaks_df.ix[refined_peaks_df.z<(np.min(refined_peaks_df.z)+dz)]
        it=0
        while len(z_layer_image_df)<3:#test next layer
            it+=1
            z_layer_image_df = refined_peaks_df.ix[refined_peaks_df.z<(np.min(refined_peaks_df.z)+it*dz)]
            z_layer_image_df = z_layer_image_df.ix[z_layer_image_df.z>=(np.min(z_layer_image_df.z)+(it-1)*dz)]
      
    if side=='right':    
        z_layer_image_df = refined_peaks_df.ix[refined_peaks_df.z>(np.max(refined_peaks_df.z)-dz/2)]
        it=0
        while len(z_layer_image_df)<3:#test next layer
            it+=1
            z_layer_image_df = refined_peaks_df.ix[refined_peaks_df.z>(np.max(refined_peaks_df.z)-it*dz)]
            z_layer_image_df = z_layer_image_df.ix[z_layer_image_df.z<=(np.max(z_layer_image_df.z)-(it-1)*dz)]
    return z_layer_image_df

def match_initial_points(template_edge_df, z_layer_image_df, match_atoms=None, atoms_im=None, atoms_temp=None):
    '''match_atoms: list of numbers in template to match with first atoms in image. 
    if 'all', no gaps in image atoms, all atoms used'''
    sorted_im = z_layer_image_df.sort_values('x_proj').copy()
    sorted_im.index=np.arange(0, len(sorted_im.index))
    sorted_tem = template_edge_df.sort_values('x_proj').copy()
    sorted_tem.index=np.arange(0, len(sorted_tem.index))
    
    if  match_atoms=='all':
        shortest=len(sorted_tem.index) if len(sorted_tem.index)<=len(sorted_im.index) else len(sorted_im.index)
        atoms_im = np.arange(0, shortest)
        atoms_temp = np.arange(0, shortest)
        
    im_points = sorted_im.ix[np.array(atoms_im)].copy()
    tem_points = sorted_tem.ix[np.array(atoms_temp)].copy()
    scale_factor_x_proj = abs(np.max(im_points.x_proj)-np.min(im_points.x_proj))/abs(np.max(tem_points.x_proj)-np.min(tem_points.x_proj))
    scaled_tem = tem_points*scale_factor_x_proj
    
    translation_x_proj = np.min(im_points.x_proj)-np.min(scaled_tem.x_proj)
    translation_z = im_points.z[np.argmin(im_points.x_proj)] - scaled_tem.z[np.argmin(scaled_tem.x_proj)] 
    
    return scale_factor_x_proj, translation_x_proj, translation_z 


def match_two_point_sets(template_df, image_df, min_dist=None, plot=False):
    '''Point pair: one in template_df and one in image_df CLOSER than smallest_dist/2 (smallest dist within template)
    returns dfs matched by index. Template input must be scaled and matched. '''
    template_array = template_df.as_matrix(columns = ['x_proj', 'z'])
    kdt = KDTree(template_array,  metric='euclidean')
    dist, inds = kdt.query([[template_array[0][0],template_array[0][1]]], k=3)  
    
    #smallest distance within template: Use half this as limit for NN    
    smallest_dist = min_dist if min_dist is not None else dist[0][1]
    
    image_array = image_df.as_matrix(columns = ['x_proj', 'z'])
    kdt = KDTree(image_array,  metric='euclidean')
    
    #iterate over all atoms in template. find nearest neighbour in the image peaks. IF clore enough, add to list.
    template_points_with_neighbour=[]
    image_points_with_neighbour=[]
    for atom_ind in template_df.index:
            dist, inds = kdt.query([[template_df.x_proj[atom_ind],template_df.z[atom_ind]]], k=1) 
            
            if dist[0][0]<smallest_dist/2:
                dict_template = {'z': template_df.z[atom_ind], 'x_proj':template_df.x_proj[atom_ind] }
                dict_image = {'z': image_array[inds[0][0]][1], 'x_proj':image_array[inds[0][0]][0]}
            
                template_points_with_neighbour.append(dict_template)
                image_points_with_neighbour.append(dict_image)
                
    template_matched_df = pd.DataFrame(template_points_with_neighbour, columns=['z', 'x_proj'])
    image_matched_df = pd.DataFrame(image_points_with_neighbour, columns=['z', 'x_proj'])
    
    colors = [plt.cm.Spectral(each)
              for each in np.linspace(0, 1, len(image_matched_df.z))]    
    random.shuffle(colors)
    
    if plot:
        fig, ax = plt.subplots()
        for i in template_matched_df.index:
            
            ax.scatter(template_matched_df.z[i], template_matched_df.x_proj[i], color=tuple(colors[i]))
            ax.scatter(image_matched_df.z[i], image_matched_df.x_proj[i], color=tuple(colors[i]))
        plt.xlabel('z')
        plt.ylabel('x_proj')
        ax.axis('equal')
        ax.set_ylim(ax.get_ylim()[::-1])  
   
    return template_matched_df, image_matched_df

def get_coeffinients(affine_matrix):
    '''calculated coefficients from affine matrix:
    
    A = np.arrar([[a0, a1, a2],
                  [b0, b1, b2]])
    translation vector t = [a2, b2] = [tx, ty]
    a0=sx cos(T)
    a1=-sy sin(T + C) #C is shear angle, T is rotation angle
    b0 = sx sin(T)
    b1 = sy sin(T + C)
    '''
    [a0, a1, a2, b0, b1, b2] =  affine_matrix.ravel()[0:6]
    
    #translation
    tx, ty = a2, b2
    #scaling
    sx = np.sqrt(a0**2 + b0**2)
    #rotation angle
    T = np.arctan(b0/a0)
    #shear angle
    C = np.arctan(-a1/b1) -T
    sy = -a1/np.sin(T+C) #=b1/np.cos(T+C)

    print('Rotation angle T: %f degrees \n'%(np.rad2deg(T)))
    print('Shear angle C (x-shear): %f degrees \n'%(np.rad2deg(C)))
    print('Scale factors [sx, sy]: \n [%f, %f] \n'%(sx, sy))
    print('Translation vector [tx, ty]: \n [%f, %f] \n'%(tx, ty))
    print('Returns [sx, sy, T, C, tx, ty] \n ')
    return [sx, sy, T, C, tx, ty]

####Perform affine transformation 
    
def estimate_affine_matrix(template_matched_df,image_matched_df):
    '''estimate affine transformation matrix. Returns matrix as numpy array'''
    src = template_matched_df.as_matrix(['z', 'x_proj'])
    dst = image_matched_df.as_matrix(['z', 'x_proj'])
    tform = transform.estimate_transform('affine', src, dst) #same! #handlar vel berre om aa faa ut rett type matrix
    print('Transformation matrix: \n')
    print(tform.params)
    affine_matrix=np.array(tform.params)
    return affine_matrix

def apply_affine_transform(affine_matrix, image):
    '''takes image as array, matrix as array, applies transformation from matrix, returns image as array'''
    tform = transform.AffineTransform(affine_matrix)
    # Normalize image to between 0 and 255
    image_for_aff = ((image-image.min())/(image.max()-image.min()))*225
    warped = transform.warp(image_for_aff, tform, preserve_range=True)#, output_shape=(50, 300))
    return warped
    