"""
### noise_reduction.py ###

Functions for image averaging by cross-correlation to reduce noise
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from skimage.feature import match_template
from scipy.signal import argrelextrema
from statsmodels.nonparametric.smoothers_lowess import lowess

import grain_boundary_reconstruction.tools as tools

#####

def generate_template(image, template_height=100):
    '''creates image_snip = image without edges
    creates template covering full width'''
    template_start_x=0  # change to ignore edges
    template_start_y=0  # change to ignore edges
    template_size=[image.shape[1]-2*template_start_x, template_height] 
    snip_size=[template_size[0], image.shape[0]]
    
    #create template
    template=image[template_start_y:(template_start_y+template_size[1]), template_start_x:(template_start_x+template_size[0])] ##using almost full width, only 1 GB unit height
    ##create image with same width as template, full height #remove edges
    image_snip=image[0:snip_size[1], template_start_x:(template_start_x+template_size[0])] 
    return image_snip, template

def cross_correlation_along_GB(image, plot=True, template_height=100, scatter_peaks=None):
    '''Creates a template within image, shifts if vertically across the image and calculates cross-correlation function (CCF).
    returns indices for maxima'''
    image_snip, template = generate_template(image, template_height=template_height)    
    CCF_GB = match_template(image_snip, template)
    #filtering to easier find peaks
    CCF_GB_filtered = lowess(CCF_GB[:,0], range(0, np.shape(CCF_GB[:,0])[0] ),frac = 0.04)[:,1]# frac=im_df.signal_filtering_frac[ind])[:,1]  
    max_ind = argrelextrema(np.array(CCF_GB_filtered), np.greater)[0] 
    
    if plot:
        fig,ax =tools.plot_image_array(image_snip)
        #frame template
        rect = patches.Rectangle((1,1),image_snip.shape[1]-3,template_height,linewidth=2,edgecolor='r',facecolor='none')
        ax.add_patch(rect)
        plt.figure()
        plt.plot(CCF_GB_filtered, color='k', linewidth=0.5) 
        plt.scatter(max_ind, [CCF_GB_filtered[ind] for ind in max_ind], color='b')
        if scatter_peaks is not None:
            plt.scatter(max_ind[scatter_peaks], [CCF_GB_filtered[ind] for ind in max_ind[scatter_peaks]], color='r', marker='*')
    return max_ind

def create_averaged_image(image, shifts, averaged_image_height):
    '''returns image averaged over units at image positions indicated by shifts'''
    image_snip, template = generate_template(image)
    image_averaged=np.zeros((averaged_image_height, image_snip.shape[1]))
    no_units=1
    for shift in shifts:
        if shift+averaged_image_height<image_snip.shape[0]: #ensure that indices are wiithin image
            template_shifted=image_snip[shift:shift+averaged_image_height, :]
            image_averaged+=template_shifted
            no_units+=1
    print('Image averaged over %d units'%no_units)    
    return image_averaged, no_units