"""
### peakFinder.py ###

Functions for finding intensity peaks and fitting 2-D Gaussian function to the data
"""
import pandas as pd
import numpy as np
from scipy import optimize

#####

def find_maxima(image, lower_intensity_limit_factor, minimum_distance):
    '''return all maxima above lower_intensity_limit_factor(factor of image max), IF distance bigger than min_dist
    atoms clores to the edge than 1 atom radius are ignored'''
    
    #add zero padding (2*atomradius) to remove problem with atoms at edges    
    image_copy = np.zeros(((image.shape[0] + minimum_distance*4), (image.shape[1] + minimum_distance*4)))
    image_copy[minimum_distance*2:-minimum_distance*2, minimum_distance*2:-minimum_distance*2] += image
    max_list=[]
    lower_intensity_limit = lower_intensity_limit_factor*np.max(image)
    y_size, x_size=image.shape
    
    peak_found=True
    while peak_found==True:
        if np.max(image_copy)>lower_intensity_limit:
            y_max, x_max = np.unravel_index(np.argmax(image_copy), image_copy.shape) # find the match   
            ##Brute force: ser area around peak to zero, find next max
            image_copy[y_max-minimum_distance:y_max+minimum_distance, x_max-minimum_distance:x_max+minimum_distance ] = 0
            #remove padding in saved coordinate
            dict1 = {'z': x_max-minimum_distance*2, 'x_proj': y_max-minimum_distance*2 }
            #if not on edge, add to list:
            if dict1['z']>minimum_distance and dict1['z']<x_size-minimum_distance:
                if dict1['x_proj']>minimum_distance and dict1['x_proj']<y_size-minimum_distance:
                    max_list.append(dict1)
        else:
            peak_found=False
           
    maxima_df = pd.DataFrame(max_list, columns=['z', 'x_proj'])    
    print("Number of atoms detected: "+str((len(maxima_df.z))))
    print("Intensity level minimum: "+str(lower_intensity_limit_factor*100)+" % of max intensity")
    return maxima_df

def gaussian(amplitude, mu_x, mu_y, sigma_x, sigma_y):
    """Returns a gaussian function with the given parameters"""
    return lambda x,y: amplitude*np.exp(
                -(((mu_x-x)/sigma_x)**2+((mu_y-y)/sigma_y)**2)/2)
    
def moments(data):
    """Returns (amplitude, x, y, sigma_x, sigma_y)
    the gaussian parameters of a 2D distribution by calculating its
    moments """
    total = data.sum()
    X, Y = np.indices(data.shape)
    x = (X*data).sum()/total
    y = (Y*data).sum()/total
    col = data[:, int(y)]
    sigma_x = np.sqrt(np.abs((np.arange(col.size)-y)**2*col).sum()/col.sum())
    row = data[int(x), :]
    sigma_y = np.sqrt(np.abs((np.arange(row.size)-x)**2*row).sum()/row.sum())
    amplitude = data.max()
    return amplitude, x, y, sigma_x, sigma_y

def fitgaussian(data):
    """Returns (amplitude, x, y, sigma_x, sigma_y)
    the gaussian parameters of a 2D distribution found by least-squares fit"""
    params = moments(data)
    errorfunc = lambda p: np.ravel(gaussian(*p)(*np.indices(data.shape)) - data)
    params_bestfit, success = optimize.leastsq(errorfunc, params)
    return params_bestfit

def fit_gaussians_to_maxima(image, initial_peaks_df, atomradius, iterations=1, intensity=False):
    '''Gaussian fit to square box of the image of size 2*atomradius around each initial peak position'''
    #add padding to avoid problems at edges
    image_copy = np.zeros(((image.shape[0] + atomradius*4), (image.shape[1] + atomradius*4)))
    image_copy[atomradius*2:-atomradius*2, atomradius*2:-atomradius*2] += image
    positions_refined_list=[]
    y_size, x_size=image.shape
    
    move_list=[]
    for ind, x,y in zip(initial_peaks_df.index,initial_peaks_df.z, initial_peaks_df.x_proj):  
        #choose area (atomradius)x(atomradius) around each peak for gaussian fitting
        atom_snip = image_copy[int(y+atomradius):int(y+atomradius*3), int(x+atomradius):int(x+atomradius*3)].copy()
        params = fitgaussian(atom_snip)
        (amplitude, y_gauss, x_gauss, sigma_x, sigma_y) = params
        move= np.sqrt((y_gauss-atomradius)**2 + (x_gauss-atomradius)**2)
        
        if move<5: #ignore those atoms that go very far(jumps out of image)
            move_list.append(move)
        if intensity:
            dict1 = {'z': x_gauss-atomradius+x, 'x_proj': y_gauss-atomradius+y, 'intensity': amplitude } #remove padding
        else:
            dict1 = {'z': x_gauss-atomradius+x, 'x_proj': y_gauss-atomradius+y } #remove padding
        #if not on edge, add to list:
        if dict1['z']>atomradius and dict1['z']<x_size-atomradius:
            if dict1['x_proj']>atomradius and dict1['x_proj']<y_size-atomradius:
                positions_refined_list.append(dict1)        
    if intensity:
        maxima_refined_df = pd.DataFrame(positions_refined_list, columns=['z', 'x_proj', 'intensity'])
    else:
        maxima_refined_df = pd.DataFrame(positions_refined_list, columns=['z', 'x_proj'])
    print("Number of atoms detected: "+str((len(maxima_refined_df.z))))
    print("Average position move: "+str(np.mean(move_list)))
    #return maxima_refined_CM_df    
    for it in range(iterations-1):
        maxima_refined_df=fit_gaussians_to_maxima(image, maxima_refined_df, atomradius, iterations=1, intensity=intensity)    
    return maxima_refined_df
