"""
### reconstruction.py ###

Functions for 3-D reconstruction from lines and image alignment
"""
import pandas as pd
import numpy as np
from sklearn.neighbors import KDTree
import matplotlib.pyplot as plt

import grain_boundary_reconstruction.tools as tools

#####

def find_atoms_from_lines(im1pos_df, im2pos_df, angles, atomradius):
    '''Takes two df med attributes x_proj, z, one for each image (ferdig aligna), and finds intersections..
    "atomradius": overlap if distance between liner are less than 2*atomradius
    angles=[a1, a2]'''
    
    A = np.array([[np.cos(np.deg2rad(angles[0])),  np.sin(np.deg2rad(angles[0]))], 
         [np.cos(np.deg2rad(angles[1])), np.sin(np.deg2rad(angles[1]))]])
    #deg2rad gir rounding errors, setter derfor lave tal til 0 her
    A[A < 7e-17] = 0
    
    
    coords_3D_list=[]
    for i in im1pos_df.index:#range(0, len(im1pos_df.z)):
        for j in im2pos_df.index:#range(0, len(im2pos_df.z)):
            z_dist = np.abs(im1pos_df.z[i]-im2pos_df.z[j])
            
            if z_dist <2*atomradius: #overlap
                z_out = (im1pos_df.z[i]+im2pos_df.z[j])/2 
                w = np.array([im1pos_df.x_proj[i], im2pos_df.x_proj[j]])
                v = np.linalg.solve(A,w)
                dict1 = {'atom': 'Zr', 'x': v[0], 'y': v[1], 'z': z_out}
                coords_3D_list.append(dict1)
                             
    pos_3D_df = pd.DataFrame(coords_3D_list, columns=['atom','x','y','z']) 
    return pos_3D_df

def find_atoms_from_lines_many_images(atompos_df_list, angle_list, atomradius, max_dist_intersections=None):
    '''Takes list of MINIMUM 2  df med attributes x_proj, z, one for each image (ferdig aligna), and finds intersections..
    "atomradius": overlap if distance between liner are less than 2*atomradius
    angles=[a1, a2, a3]
    Uses e.find_atoms_from_lines and e.cluster '''
    max_dist_intersections = max_dist_intersections if max_dist_intersections is not None else atomradius
    
    no_df = len(atompos_df_list) #minimum2 !
    
    pairs_list = []
    
    #i,j=[0,1]
    ##no_df = 2
    for i in range(0, no_df):
        for j in range(i, no_df):            
            if i!=j:
                print('Images combined: '+str(i)+' and '+str(j))
            
                pos_3D_df_pair = find_atoms_from_lines(atompos_df_list[i], atompos_df_list[j], [angle_list[i],angle_list[j]], atomradius)
                print('Number of atoms found: '+str(len(pos_3D_df_pair)))
                pairs_list.append(pos_3D_df_pair)
    
    #combine
    all_df = pd.concat(pairs_list)
    
    #3D clustering
    if no_df>2:
        minsample=no_df #cluster: find mean if three intersections are close...
        pos_3D_df = tools.cluster_3D(all_df, minsample, max_dist=max_dist_intersections, maxsize=no_df)
        #pos_3D_df = cluster(all_df, minsample, max_dist=max_dist_intersections)
        #tools.plot3Dstructure(pos_3D_df)
    else:
        pos_3D_df=all_df
    return pos_3D_df

def align_image_to_reprojection(reproj_cluster_df, image_df, interval, dx, d_NN_proj_row, plot=True):
    '''Finds how much image_df should be moved to minimize mean dist to reprojected image atoms'''
    #check within each layer only. and only if nn is less than dv/2 to ensure that not next row nn is used (if atoms are missing)
    
    ##IF we make sure the interval is less than d_NN_proj: can use NN in combined image with both reproj and image_adjust!
    #for each atom in image+adjust, check if it has NN 
    
    image_adjusted_df = image_df.copy()
    image_adjusted_df_bulk = pd.concat([image_adjusted_df[image_adjusted_df.z<-0.5], image_adjusted_df[image_adjusted_df.z>2.5]])

    image_adjusted_df_bulk.x_proj+=interval[0] #set image to first test position
    
    atompos_array = reproj_cluster_df.as_matrix(columns = ['x_proj', 'z'])
    
    mean_list=[]
    for adj in np.arange(interval[0], interval[1], dx):
        image_adjusted_df_bulk.x_proj+=dx #move image slowly upwards
        
        NN_dist_list=[]
        for atom_ind in image_adjusted_df_bulk.index:
            kdt = KDTree(atompos_array,  metric='euclidean')
            dist, inds = kdt.query([[image_adjusted_df_bulk.x_proj[atom_ind],image_adjusted_df_bulk.z[atom_ind]]], k=1)  
            #dist, inds = kdt.query([[0,0]], k=1)  
            
            NN_dist_list.append(dist[0][0])
        NN_dist_array=np.array(NN_dist_list)
        mean_list.append(np.mean(NN_dist_array[NN_dist_array<d_NN_proj_row])) #just to remove the edge atoms...       
        
    if plot:
        plt.figure()
        plt.scatter(np.arange(interval[0], interval[1], dx),mean_list)
    
    minimizing_adj = np.arange(interval[0], interval[1], dx)[np.argmin(mean_list)]

    print('Image shift minimizing mean distance to nearest neighbour: %.2f'%(minimizing_adj))
    print('Mean distance to nearest neighbour: %.2f'%(np.min(mean_list)))
    
    if plot:
        plt.figure()
        plt.scatter(reproj_cluster_df.z, reproj_cluster_df.x_proj, s=10, color='b')
        plt.scatter(image_adjusted_df.z, image_adjusted_df.x_proj+minimizing_adj, s=5, color='r')
        plt.title('Image shifted %.2f vertically'%(minimizing_adj))
    return minimizing_adj
