"""
### test_data.py ###

Functions for retrieving data from the folder test_data_folder. Files were stored here for each step in the reconstruction process. Current directory should be set within the grain_boundary_reconstruction folder. If not, following line should be changed in get-functions:
path_dir = os.getcwd()
"""
from skimage import io
import os
import matplotlib.pyplot as plt
from scipy import ndimage
import pandas as pd
import numpy as np

import grain_boundary_reconstruction.tools as tools

#####

def get_original_images():
    '''returns list of arrays containing original images'''
    path_dir = os.getcwd()
    path_images=r'test_data_folder\\original_images_sigma5'
    filename_list = os.listdir(path_dir+'\\'+path_images)
    images_list = [io.imread(path_dir+'\\'+path_images+'\\'+filename, as_grey=True) for filename in filename_list]
    return images_list

def get_refined_peaks_original():
    '''returns list of df's containing refined peak positions for averaged images.
    This is the starting point for the 3-D reconstruction.'''
    path_dir = os.getcwd()
    path_peaks=r'test_data_folder\\\original_images_atompos'
    filename_list = os.listdir(path_dir+'\\'+path_peaks)
    peaks_list = [pd.DataFrame(np.load(path_dir+'\\'+path_peaks+'\\'+filename), columns=['z', 'x_proj']) for filename in filename_list]
    return peaks_list

def get_affined_images():
    '''returns list of arrays containing affined images'''
    path_dir = os.getcwd()
    path_images=r'affined_images'
    images_list=[]
    for i in range(0,5):
        filename='affined_avg_img'+str(i)+'.png'
        image_arr=io.imread(path_dir+'\\'+path_images+'\\'+filename, as_grey=True)
        images_list.append(np.array(image_arr, dtype=float))
    return images_list

def get_averaged_images():
    '''returns list of arrays containing averaged images'''
    path_dir = os.getcwd()
    path_images=r'test_data_folder\\averaged_images'
    filename_list = os.listdir(path_dir+'\\'+path_images)
    images_list = [io.imread(path_dir+'\\'+path_images+'\\'+filename, as_grey=True) for filename in filename_list]
    return images_list

def get_refined_peaks_avg():
    '''returns list of df's containing refined peak positions for averaged images.
    This is the starting point for the 3-D reconstruction.'''    
    path_dir = os.getcwd() 
    path_peaks=r'test_data_folder\\averaged_images_atompos'
    filename_list = os.listdir(path_dir+'\\'+path_peaks)
    peaks_list = [pd.DataFrame(np.load(path_dir+'\\'+path_peaks+'\\'+filename), columns=['z', 'x_proj']) for filename in filename_list]
    return peaks_list

def get_refined_peaks_avg_scaled():
    '''returns list of df's containing refined peak positions for averaged images.
    This is the starting point for the 3-D reconstruction.'''
    path_dir = os.getcwd()
    path_peaks=r'test_data_folder\\averaged_images_atompos_scaled'
    filename_list = os.listdir(path_dir+'\\'+path_peaks)
    peaks_list = [pd.DataFrame(np.load(path_dir+'\\'+path_peaks+'\\'+filename), columns=['z', 'x_proj']) for filename in filename_list]
    return peaks_list

def get_reconstructed_from_cif(image_inds, n_units=1):
    '''returns the structure'rec. from image inds from cif-file'''
    images_ind_str = '_images_'
    for ind in image_inds:
        images_ind_str+=(str(ind)+'_')
        
    path_dir = os.getcwd()
    path_files=r'reconstructions'    
    filename_list = os.listdir(path_dir+'\\'+path_files)
    check=0
    #find filename that contains image_inds_str    
    filename_start = 'Reconstruction_REAL_'+images_ind_str+'atomradius_'
    for filename in filename_list:
        if filename[0:len(filename_start)]==filename_start:  
            if filename[-3:]=='cif':
                df_cif = tools.readCIF(path_dir+'\\'+path_files+'\\'+filename[:-4], n_units=n_units)
                check=1
    if check==0:
        print('File not found!')
        df_cif=0
    return df_cif

def get_reconstructed_from_xyz(image_inds):
    '''returns the structure'rec. from image inds from cif-file'''
    images_ind_str = '_images_'
    for ind in image_inds:
        images_ind_str+=(str(ind)+'_')
    path_dir = os.getcwd() 
    path_files=r'reconstructions'    
    filename_list = os.listdir(path_dir+'\\'+path_files)
    check=0
    #find filename that contains image_inds_str    
    filename_start = 'Reconstruction_REAL_'+images_ind_str+'atomradius_'
    for filename in filename_list:
        if filename[0:len(filename_start)]==filename_start:  
            if filename[-3:]=='xyz':
                df_from_xyz = tools.readXYZ(path_dir+'\\'+path_files+'\\'+filename[:-4])
                check=1
    if check==0:
        print('File not found!')    
    return df_from_xyz

def get_ideal_bulk():
    '''Returns ideal bulk'''    
    pd.options.mode.chained_assignment = None  # default='warn' #get rid of warning
    path=r'test_data_folder\\ideal_3D_structure'
    filename_ideal='half_sigma5.xyz'
    ideal_df = pd.read_csv(path+'\\'+filename_ideal, sep=" ", header=1, names = ['atom', 'x', 'y', 'z'])
    ideal_out = ideal_df[['x', 'y', 'z']]
    return ideal_out
    
def get_reconstructed_structure(n_units=3):
    '''Returns reconstructed structure as multiples of real clustered CIF-file...''' 
    filename='real_rec_'+str(n_units)+'units.xyz'
    path='test_data_folder'
    return pd.read_csv(path+'\\'+filename, delimiter=' ', header=2, names=['atom', 'x', 'y', 'z'])

def get_angles_deg():
    '''Calculate rotation angles from lattice directions'''
    dirZone =  np.array([0, 0, 1]) 
    dir38 =  np.array([1, -3, 4]) 
    dir58 =  np.array([1, -3, 2]) 
    dir90 =  np.array([1, -3, 0]) 
    theta38 = np.rad2deg(np.arccos( (np.sum(dirZone*dir38))/( np.sqrt(np.sum(dirZone**2))* np.sqrt(np.sum(dir38**2)) ) ))
    theta58 = np.rad2deg(np.arccos( (np.sum(dirZone*dir58))/( np.sqrt(np.sum(dirZone**2))* np.sqrt(np.sum(dir58**2)) ) ))  
    angles_deg=[0.0, theta38, theta58, 90.0]
    return angles_deg

def get_EDX_images():
    '''returns two lists of arrays containing original EDX images,
    Zr and Y'''
    path_images=r'original_EDX_images_sigma5' ##should be in grain_boundar_reconstruction folder!!!
    names_Zr =['zone_Zr K', '38_Zr K', '58_Zr K', '90_Zr K']
    names_Y =['zone_Y K', '38_Y K',  '58_Y K',  '90_Y K']
    EDX_df_list_Zr = []
    EDX_df_list_Y = []
    for i in range(0,len(names_Zr)):
        name=names_Zr[i]
        EDX_df =  pd.read_csv(path_images+'\\'+name+'.csv', sep=",", header=None) 
        EDX_df_list_Zr.append(EDX_df)
        name=names_Y[i]
        EDX_df =  pd.read_csv(path_images+'\\'+name+'.csv', sep=",", header=None) 
        EDX_df_list_Y.append(EDX_df)
    return EDX_df_list_Zr, EDX_df_list_Y

def plot_EDX_subplot(filter_factor=0):
    EDX_df_list_Zr, EDX_df_list_Y = get_EDX_images()
    
    fig, axs = plt.subplots(4,3, figsize=(15, 6))
    fig.subplots_adjust(top=0.97,
                        bottom=0.0,
                        left=0.0,
                        right=1.0,
                        hspace=0.085,
                        wspace=0.0)
    X,Y = np.meshgrid(range(0,EDX_df_list_Zr[0].shape[0]),range(0,EDX_df_list_Zr[0].shape[1]))
    axs = axs.ravel()
    ind=-1
    for i in [0,3,6,9]:
        ind+=1
        EDX_df_both = EDX_df_list_Zr[ind] + EDX_df_list_Y[ind]  
        intensity_both = ndimage.filters.gaussian_filter(np.asarray(EDX_df_both), filter_factor)
        axs[i].pcolormesh(X,Y,intensity_both)#, cmap='Greys_r', vmin=vmin, vmax=vmax)
        axs[i].set_title("Sum Y and Zr image "+str(ind))
        axs[i].axis('equal')
        axs[i].axis('off')
        EDX_df_Zr = EDX_df_list_Zr[ind] 
        intensity_Zr = ndimage.filters.gaussian_filter(np.asarray(EDX_df_Zr), filter_factor)
        axs[i+1].pcolormesh(X,Y,intensity_Zr)#, cmap='Greys_r', vmin=vmin, vmax=vmax)
        axs[i+1].set_title("Zr image "+str(ind))
        axs[i+1].axis('equal')
        axs[i+1].axis('off')
        EDX_df_Y = EDX_df_list_Y[ind]  
        intensity_Y = ndimage.filters.gaussian_filter(np.asarray(EDX_df_Y), filter_factor)
        axs[i+2].pcolormesh(X,Y,intensity_Y)#, cmap='Greys_r', vmin=vmin, vmax=vmax)
        axs[i+2].set_title("Y image "+str(ind))
        axs[i+2].axis('equal')
        axs[i+2].axis('off')
            
def get_visual_guess_GB_z_list(average=False):   
    '''returns list of visual GB positions'''
    if average:
        ut = np.array([[194, 214],[405, 450], [467, 511], [248, 262]])
    else:
        ut = np.array([[198, 214],[405, 450], [473, 520], [248, 262]])
    return ut

def refine_GB_z_values_and_find_x_proj_unit_dist(atompos_df_list, visual_guess_GB_z_list, dz=5, plot=False):
    '''Takes initial guess for GB position, returns mean z-values of the atom layers indicated'''
    
    visual_guess_GB_z_list_refined = []
    unit_x_proj_dist_list=[]
    for i in range(0,4):
        pos_df = atompos_df_list[i]
        GB_L_z = visual_guess_GB_z_list[i][0]
        GB_R_z = visual_guess_GB_z_list[i][1]
        GB_L_inds =[np.array(np.array([pos_df.z<(GB_L_z+dz) ])*np.array([pos_df.z>(GB_L_z-dz) ]))][0][0] 
        GB_R_inds =[np.array(np.array([pos_df.z<(GB_R_z+dz) ])*np.array([pos_df.z>(GB_R_z-dz) ]))][0][0] 
        GB_L_z_refined = np.mean(pos_df.z[GB_L_inds])
        GB_R_z_refined = np.mean(pos_df.z[GB_R_inds])
        
        visual_guess_GB_z_list_refined.append([GB_L_z_refined, GB_R_z_refined])
        if plot:
            plt.figure()
            plt.scatter(pos_df.z, pos_df.x_proj,  s=10, color='b', label='bulk atoms')
            plt.scatter(pos_df.z[GB_L_inds], pos_df.x_proj[GB_L_inds],  s=4, color='r', label='left GB layer')
            plt.scatter(pos_df.z[GB_R_inds], pos_df.x_proj[GB_R_inds],  s=4, color='y', label='right GB layer')
            plt.axis('equal')
            plt.legend()
            
        #Find distance between GB units
        x_vals = sorted(np.array(pos_df.x_proj[GB_L_inds]))
        
        if len(x_vals)<2: 
            print('Too few GB atoms found')
        else:
            GB_unit_2_atom_x=x_vals[0]
            if i==0:
                GB_unit_1_atom_x=x_vals[2]   
            if i==1:  
                GB_unit_1_atom_x=x_vals[1]
            if i==2:  
                GB_unit_1_atom_x=x_vals[1]
            if i==3:  
                GB_unit_1_atom_x=x_vals[3]
            
            unit_x_proj_dist = abs(GB_unit_2_atom_x-GB_unit_1_atom_x) 
            unit_x_proj_dist_list.append(unit_x_proj_dist)
    return visual_guess_GB_z_list_refined, unit_x_proj_dist_list

def get_scale_factors():
    '''returns scale factors (px-to-angstrom ratio) for affined images (should be same as for averaged images). '''
    path_dir = os.getcwd() 
    path=r'test_data_folder'
    scale_factors = np.load(path_dir+'\\'+path+'\\scale_factors_avg.npy')
    return scale_factors