"""
### tools.py ###

Tools for plotting, reading and writing files, kernel density estimation, NN distances, matching point sets for RMSD estimation etc.
"""
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from mpl_toolkits.mplot3d import proj3d
from sklearn.cluster import DBSCAN
import time
from scipy import ndimage
from sklearn.neighbors import KDTree
from sklearn.neighbors import KernelDensity
import random

#####

def plot_image_array(image, axis=True, filter_factor=0, title=False, figHeight=5, cmap='Greys_r'):
    '''Plot image array'''    
    x=np.arange(0, np.shape(image)[1])
    y=np.arange(0, np.shape(image)[0])
    X,Y=np.meshgrid(x,y)
    imratio=image.shape[0]/image.shape[1]
    fig,ax = plt.subplots(1, figsize=(figHeight, imratio*figHeight))
    image_filtered= ndimage.filters.gaussian_filter(np.asarray(image), filter_factor)
    ax.pcolormesh(X,Y,image_filtered, cmap=cmap)
    plt.axis('equal')
    plt.subplots_adjust(left=0.12,  right=0.99, bottom=0.12, top=0.99, wspace=0, hspace=0)
    plt.xlabel('$z$')
    plt.ylabel('$x_p$')
    
    if axis is not True:
        plt.axis('off')
        plt.subplots_adjust(left=0,  right=1, bottom=0, top=1, wspace=0, hspace=0)
        
    if title is not False:
        plt.subplots_adjust(left=0.12,  right=0.99, bottom=0.12, top=0.9, wspace=0, hspace=0)
        plt.title(title)
    return fig, ax
        
def plot_2D_df(projection_2D_df, withAxis=True, figSize=(4,4), markerSize  = 10, color='b'):
    '''Plot 2D reprojection'''
    fig, ax = plt.subplots(figsize=figSize)
    ax.scatter(projection_2D_df.z, projection_2D_df.x_proj,  s=markerSize, color=color)
    ax.axis('equal')
    #ax.set_ylim(ax.get_ylim()[::-1])  
    
    if withAxis:
        ax.set_xlabel('z')
        ax.set_ylabel('x_proj')
        fig.subplots_adjust(left=0.16,  right=0.98, bottom=0.11, top=0.98, wspace=0, hspace=0)
    else:
        ax.axis('off')
        fig.subplots_adjust(left=0,  right=1, bottom=0, top=1, wspace=0, hspace=0)

def plot3Dstructure(atompos_df, df_2=None, markerSize=10, view_dir=0, orthogonal=True,linewidths=0.5,elev=0, panes=False, figSize = (6,6), xlim=None, ylim =None, zlim=None, alpha=0.7, colors=None, c=[0,0,1], c2=[1,0,0]):
    xlim = xlim or [np.min(atompos_df.x),np.max(atompos_df.x)]
    ylim = ylim or [np.min(atompos_df.y),np.max(atompos_df.y)]
    zlim = zlim or [np.min(atompos_df.z),np.max(atompos_df.z)] #bestemmer utifraa df size OR arguments

    def axisEqual3D(ax):
        extents = np.array([getattr(ax, 'get_{}lim'.format(dim))() for dim in 'xyz'])
        sz = extents[:,1] - extents[:,0]
        centers = np.mean(extents, axis=1)
        maxsize = max(abs(sz))
        r = maxsize/2
        for ctr, dim in zip(centers, 'xyz'):
            getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)
              
    if orthogonal:       
        def orthogonal_proj(zfront, zback):
            a = (zfront+zback)/(zfront-zback)
            b = -2*(zfront*zback)/(zfront-zback)
            return np.array([[1,0,0,0],
                                [0,1,0,0],
                                [0,0,a,b],
                                [0,0,0,zback]])
        
        proj3d.persp_transformation = orthogonal_proj 
        
    if  colors is not None:
        if colors=='pair':
            colors = [plt.cm.Spectral(each) for each in np.linspace(0, 1, len(atompos_df.z))]    
            random.shuffle(colors)
        rgba_colors=colors
        rgba_colors_2=colors
        
    else:
        rgba_colors = np.zeros((atompos_df.shape[0],4))
        rgba_colors[:,0] = c[0] #red
        rgba_colors[:,1] = c[1] #green
        rgba_colors[:,2] = c[2] #blue
        alphas =np.ones(atompos_df.shape[0])*alpha
        rgba_colors[:, 3] = alphas
        
        if df_2 is not None:
            if colors is None: 
                rgba_colors_2 = np.zeros((df_2.shape[0],4))
                rgba_colors_2[:,0] = c2[0] #red
                rgba_colors_2[:,1] = c2[1] #green
                rgba_colors_2[:,2] = c2[2] #blue
                alphas =np.ones(df_2.shape[0])*alpha
                rgba_colors_2[:, 3] = alphas

    fig = plt.figure(figsize=figSize)
    ax = fig.add_subplot(111, projection='3d')
    ax.view_init(elev=elev, azim=270+view_dir) 
    ax.scatter(atompos_df.x,
               atompos_df.y,
               atompos_df.z,
               s=markerSize, marker='o',c=rgba_colors, edgecolors='k',linewidths=linewidths, zdir='z')
    
    if df_2 is not None:
        ax.scatter(df_2.x,
                   df_2.y,
                   df_2.z,
                   s=markerSize, marker='o',c=rgba_colors_2, edgecolors='r',linewidths=linewidths, zdir='z')
        
    #####Layout####
    ax.set_xlim3d(xlim)
    ax.set_ylim3d(ylim)
    ax.set_zlim3d(zlim)
    axisEqual3D(ax)
    plt.subplots_adjust(left=0,  right=1, bottom=0, top=1, wspace=0, hspace=0)
    
    if panes==False:
        #Get rid of panes
        ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        # Get rid of the spines
        ax.w_xaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.w_yaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.w_zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        # Get rid of the ticks
        ax.set_xticks([]) 
        ax.set_yticks([]) 
        ax.set_zticks([])
    else:
        ax.set_xlabel('$x$') 
        ax.set_ylabel('$y$') 
        ax.set_zlabel('$z$')
    return fig, ax
    
def find_projected_coordinates(atompos_df, proj_angle, side_tilt=0):
    ''' Takes df with list of 3D coordinates (x_in, y_in, z_in) where z axis is normal to GB plane
    Returns df with 2D coordinates (x_proj, z_in), a projection in wanted direction proj_angle (input in degrees)'''
    
    #projecting matrix
    theta = np.deg2rad(proj_angle)
    #Projection matrix
    P = np.array([[np.cos(theta), np.sin(theta), 0], [0, 0, 1]])
    # using equation  x_proj = x \cos \theta + y \sin \theta. z=z.

    #tilt about x_proj axis - matrix
    cos=np.cos(np.deg2rad(side_tilt))
    sin=np.sin(np.deg2rad(side_tilt))
    ux=np.cos(theta)
    uy=np.sin(theta)
    uz=0
    
    R_x_proj = np.array([[cos+ux*ux*(1-cos),ux*uy*(1-cos)-uz*sin, ux*uz*(1-cos)+uy*sin],
                   [ux*uy*(1-cos)-uz*sin,cos+uy*uy*(1-cos), uy*uz*(1-cos)+ux*sin],
                   [ux*uz*(1-cos)-uy*sin, ux*uz*(1-cos)+ux*sin, cos+uz*uz*(1-cos)]])
    
    xyz = np.array(atompos_df.as_matrix(['x', 'y', 'z'])).T
    
    #rotate
    x,y,z=R_x_proj @ xyz
    #project
    x_proj, z = P @ np.array([x,y,z])
    projection_df = pd.DataFrame({'x_proj': x_proj, 'z': z}) 
    return projection_df

def get_column(structure_df, coord_z, coord_x_proj, angle_deg, radius=0.2):
    '''returns df with atoms in the column specified in coords=[x_proj,z] in projectionimage'''
    a=5.17
    dz=a/(2*np.sqrt(10))    
    #get right layer
    column_df = structure_df[structure_df.z<coord_z+dz/2].copy()
    column_df =column_df.drop(column_df.index[column_df.z<coord_z-dz/2])
    
    #test alle atom i this layer, sjekk om dei er naert nok using equation  x_proj = x \cos \theta + y \sin \theta
    for ind in column_df.index:
        x=column_df.x[ind]
        y=column_df.y[ind]
        x_proj = x*np.cos(np.deg2rad(angle_deg)) + y*np.sin(np.deg2rad(angle_deg))
        
        if abs(x_proj-coord_x_proj)>radius:
            column_df=column_df.drop(ind)
    return column_df

def cluster_2D_projection(atompos_proj_df, max_dist): #atompos_proj_df has x_proj, z-coordinates 
    '''Performs clustering on 2D projected df, 
    returns center_df with sixes ad labe'''
    
    time_start=time.time()
    z = np.array(atompos_proj_df.z)
    x_proj = np.array(atompos_proj_df.x_proj)
    
    X2d = np.array([ [z[i], x_proj[i]] for i in range(0, len(z)) ])
    db = DBSCAN(eps=max_dist, min_samples=1).fit(X2d) 
    
    labels=db.labels_
    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    
    ##Reknar ut center som CM av kvar cluster..
    center_list = []    
    for i in range(0, n_clusters_):
        atom=X2d[labels==i]
        sizes = np.sum(labels==i)
        x_proj_mean = np.mean(atom[:,1])
        z_mean = np.mean(atom[:,0])
        dict1 = {'x_proj': x_proj_mean, 'z': z_mean, 'sizes': sizes, 'labe': i }
        center_list.append(dict1)
        
    center_db_df = pd.DataFrame(center_list, columns=['x_proj', 'z', 'sizes', 'labe'])
    print("Found %d clusters \n Clustering on %d points \n Time: %f minutes \n "  %(n_clusters_, X2d.shape[0],(time.time()-time_start)/60  ))
    return center_db_df
    
def label_2D_projection(atompos_proj_df, max_dist):
    '''Performs clustering on 2D projected df, 
    returns copy of df in with labels labe and sizes of cluster'''

    z = np.array(atompos_proj_df.z)
    x_proj = np.array(atompos_proj_df.x_proj)
    
    X2d = np.array([ [z[i], x_proj[i]] for i in range(0, len(z)) ])
    db = DBSCAN(eps=max_dist, min_samples=1).fit(X2d) 
    
    labels=db.labels_      
    
    sizes_list = []    
    for i in range(0, len(labels)):
        sizes = np.sum(labels==labels[i])
        sizes_list.append(sizes)
    
    atompos_proj_df_with_labels=atompos_proj_df.copy()
    atompos_proj_df_with_labels['labe']=labels
    atompos_proj_df_with_labels['sizes']=sizes_list
    return atompos_proj_df_with_labels

def cluster_3D(overlap_df, minsample=1, max_dist = 2): 
    '''Returns DataFrame with clusters found by DBSCAN'''
    
    X3d = np.array([np.array(overlap_df.z), np.array(overlap_df.y), np.array(overlap_df.x)]).T
    db = DBSCAN(eps=max_dist, min_samples=minsample).fit(X3d)  ###kan evt justere min_samples til mykje hoegare! max size rundt 80 (90) paa atoma
    
    labels=db.labels_
    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
   
    ##Reknar ut center som CM av kvar cluster
    center_list = []    
    for i in range(0, n_clusters_):
        atom=X3d[labels==i]
        sizes = np.sum(labels==i)
        z_mean = np.mean(atom[:,0])
        y_mean = np.mean(atom[:,1])
        x_mean = np.mean(atom[:,2])
        dict1 = {'atom': 'Zr', 'x': x_mean, 'y': y_mean, 'z': z_mean, 'sizes': sizes }
        center_list.append(dict1)
        
    center_db_df = pd.DataFrame(center_list, columns=['atom','x', 'y', 'z', 'sizes'])
    print("Found %d clusters \n Clustering on %d points \n Minsample= %d \n "  %(n_clusters_, overlap_df.z.size, minsample  ))
    return center_db_df 

def printXYZ(center_db_df, filename, sizes=True, header=None): #tek inn df med x, y, z, size. filename uten extention
    '''Takes filename without extention. creates df with x,y,z,sizes entries, if sizes=true... . '''
    #heller lage func for aa lage xyz om til crystalmakervennlege filer (utan sizes) seinare
    file = open(filename+'.xyz','w') 
    
    no_atoms=len(center_db_df.z)
    file.write(str(no_atoms) + "\n")  #numbers of atoms
    
    #to avoid index mistakes
    xs=np.array(center_db_df.x)
    ys=np.array(center_db_df.y)
    zs=np.array(center_db_df.z)
    atoms=np.array(center_db_df.atom)
    
    if sizes:
        header = header if header is not None else "Reconstruction, xyz-file with size"
        file.write(header+"\n")
        col_names_line = "atom x y z sizes"
        file.write(col_names_line+"\n")
        sizes=np.array(center_db_df.sizes)
        
        for i in range(no_atoms):
            pos_line = atoms[i]+" %f %f %f %d\n"  %(xs[i], ys[i], zs[i], sizes[i])
            file.write(pos_line)         
    else:
        header = header if header is not None else  "Reconstruction, xyz-file without size"       
        file.write(header+"\n")
        col_names_line = "atom x y z"
        file.write(col_names_line+"\n")
        for i in range(no_atoms):
            pos_line = atoms[i]+" %f %f %f\n"  %(xs[i], ys[i], zs[i])
            file.write(pos_line) 
            
    file.close() 

def readXYZ(name, path=None, print_col_names=True):
    '''Takes filename without extentions, with atomname x y z and maybe size.'''
    
    filename = path+'/'+name if path is not None else name
        
    f = open(filename+'.xyz', 'r')
    for i in range(3):
        line = f.readline()
    f.close()    
    
    if len(line.split(' ')) == 5:
         atompos_df = pd.read_csv(filename+'.xyz', sep=" ", header=2, index_col=False, names = ['atom', 'x', 'y', 'z', 'sizes']) 
    elif len(line.split(' ')) == 4:
        atompos_df = pd.read_csv(filename+'.xyz', sep=" ", header=2, index_col=False, names = ['atom', 'x', 'y', 'z']) 
    else:
        print('Check file format')
        atompos_df = 0
    if print_col_names:
        print(atompos_df.columns)
    return atompos_df

def printCIF(filename, ideal_df, cell_length_a, cell_length_b, cell_length_c ):
    '''Takes xyz, transform to cif files with fraction coordinates.'''
    x_min = np.min(ideal_df.x)
    y_min = np.min(ideal_df.y)
    z_min = np.min(ideal_df.z)
    angle=90.0
    
    file = open(filename+'.cif','w') 
    file.write("data_cif\n")
    file.write("_cell_length_a\t%f\n"%(cell_length_a))
    file.write("_cell_length_b\t%f\n"%(cell_length_b))
    file.write("_cell_length_c\t%f\n"%(cell_length_c))
    file.write("_cell_angle_alpha\t%f\n"%(angle))
    file.write("_cell_angle_beta\t%f\n"%(angle))
    file.write("_cell_angle_gamma\t%f\n"%(angle))
    file.write("loop_\n")
    file.write("_atom_site_label\n")
    file.write("_atom_site_fract_x\n")
    file.write("_atom_site_fract_y\n")
    file.write("_atom_site_fract_z\n")
    
    atoms=np.array(ideal_df.atom)
    xs=np.array(ideal_df.x)
    ys=np.array(ideal_df.y)
    zs=np.array(ideal_df.z)
    
    for i in range(len(ideal_df.index)):
        file.write(atoms[i]+"\t%f\t%f\t%f\n"%((xs[i]-x_min)/cell_length_a,
                                       (ys[i]-y_min)/cell_length_b,
                                       (zs[i]-z_min)/(cell_length_c)))
    file.close() 
    print("File created:\n"+filename)
        
def clusterCIF(atompos_df, cell_length_a, cell_length_b, cell_length_c, minsample=1, max_dist=0.0001):
    '''takes big df, creates unit with specified cell lengths. This gives many atoms in the same spot. Does clustering before printing to cif-file.'''
    
    real_rec_fract_df = atompos_df.copy()
    x_min=np.min(real_rec_fract_df.x)
    y_min=np.min(real_rec_fract_df.y)
    z_min=np.min(real_rec_fract_df.z)
    
    real_rec_fract_df.x = (real_rec_fract_df.x-x_min)/(cell_length_a)
    real_rec_fract_df.y = (real_rec_fract_df.y-y_min)/(cell_length_b)
    real_rec_fract_df.z = (real_rec_fract_df.z-z_min)/(cell_length_c) #just to avoid weird edges
    bit_df=real_rec_fract_df.copy()
    
    n_x_units=int(np.max(real_rec_fract_df.x)+1)
    n_y_units=int(np.max(real_rec_fract_df.y)+1)
    
    for i in range(1, n_x_units):
        unit_inds = [np.array(np.array([real_rec_fract_df.x<(i+1)])*np.array([real_rec_fract_df.x>=i]))][0][0] 
        bit_df.x[unit_inds] = real_rec_fract_df.x[unit_inds]-i
        
    for i in range(1, n_y_units):   
        unit_inds = [np.array(np.array([real_rec_fract_df.y<(i+1)])*np.array([real_rec_fract_df.y>=i]))][0][0] 
        bit_df.y[unit_inds] = real_rec_fract_df.y[unit_inds]-i
    
    real_rec_clustered_df = cluster_3D(bit_df, minsample=minsample, max_dist=max_dist)
    
    real_rec_clustered_df.x=(real_rec_clustered_df.x*(cell_length_a)+x_min)
    real_rec_clustered_df.y=(real_rec_clustered_df.y*(cell_length_b)+y_min)
    real_rec_clustered_df.z=(real_rec_clustered_df.z*(cell_length_c)+z_min)
    return real_rec_clustered_df

def readCIF(filename, n_units=1):
    '''takes big df, creates unit with specified cell lengths. This gives many atoms in the same spot. Does clustering before printing to cif-file.'''
    
    f = open(filename+'.cif', 'r')    
    line = f.readline()
    line = f.readline()
    cell_length_x=float(line.split('\t')[1][:-1])
    line = f.readline()
    cell_length_y=float(line.split('\t')[1][:-1])
    line = f.readline()
    cell_length_z=float(line.split('\t')[1][:-1])    
    f.close()  
        
    atompos_df = pd.read_csv(filename+'.cif', sep="\t", header=11, index_col=False, names = ['atom', 'x', 'y', 'z']) 
    atompos_df.x*=cell_length_x
    atompos_df.y*=cell_length_y
    atompos_df.z*=cell_length_z

    #multiply
    unit_df_big=atompos_df.copy()
    for i in range(0, n_units):
        for j in range(0, n_units):
            shifted_unit_df=atompos_df.copy()
            shifted_unit_df.x+=i*cell_length_x
            shifted_unit_df.y+=j*cell_length_y
            
            unit_df_big=pd.concat([unit_df_big, shifted_unit_df])
        
    #remove overlapping atoms
    minsample=1 #cluster: find mean if three intersections are close.
    unit_df_big_clustered = cluster_3D(unit_df_big, minsample, max_dist=0.01)
    return unit_df_big_clustered

def rotate_3D(atompos_df, rotation_angle_deg=90, center_x=None, center_z=None):
    '''Tilting crystal. Takes angle (in deg) atompos.df with columns x, y, z, and rotates about center of df (rotation axis y)
    z_r = z_0 \cos \theta_r - x_0 \sin \theta_r
    x_r = x_0 \cos \theta_r + z_0 \sin \theta_r
    $z_0, x_0$ is input coordinates and $\theta_r$ is the rotation angle about the center of the image. 
    '''
    atompo_df=atompos_df.copy()
    
    #shift origin to center of image    
    center_x = center_x if center_x is not None else np.mean(atompo_df.x)
    center_z = center_z if center_z is not None else np.mean(atompo_df.z)
    
    atompo_df.x=atompo_df.x-center_x
    atompo_df.z=atompo_df.z-center_z
    
    theta_rotate = np.deg2rad(rotation_angle_deg) 
    
    A = np.array([[np.cos(theta_rotate),  -np.sin(theta_rotate)], 
         [np.sin(theta_rotate), np.cos(theta_rotate)]])
        
    #calculated rotated point coordinates 
    w = np.array([atompo_df.z, atompo_df.x])
    w_rotated = np.linalg.solve(A,w)
    #shift origin back
    atompos_rotated_df = pd.DataFrame({ 'x': w_rotated[1]+center_x, 'y':np.array(atompo_df.y), 'z': w_rotated[0]+center_z, })
    return atompos_rotated_df

def match_two_point_sets_3D(template_df, image_df, plot=False):
    '''Point pair: one in template_df and one in image_df CLOSER than smallest_dist/2 (smallest dist within template)
    returns dfs matched by index. Template input must be scaled and matched. '''

    template_array = template_df.as_matrix(columns = ['x', 'y', 'z'])
    kdt = KDTree(template_array,  metric='euclidean')
    dist, inds = kdt.query([[template_array[0][0],template_array[0][1],template_array[0][2]]], k=3)  
    #smallest distance within template: Use half this as limit for NN
    smallest_dist = dist[0][1]
    image_array = image_df.as_matrix(columns = ['x','y', 'z'])
    kdt = KDTree(image_array,  metric='euclidean')
    
    #iterate over all atoms in template. find nearest neighbour in the image peaks. IF clore enough, add to list.
    template_points_with_neighbour=[]
    image_points_with_neighbour=[]
    for atom_ind in template_df.index:
            dist, inds = kdt.query([[template_df.x[atom_ind], template_df.y[atom_ind],template_df.z[atom_ind]]], k=1) 
            
            if dist[0][0]<smallest_dist/2:
                dict_template = {'x':template_df.x[atom_ind], 'y':template_df.y[atom_ind] ,'z': template_df.z[atom_ind] }
                dict_image = { 'x':image_array[inds[0][0]][0],  'y':image_array[inds[0][0]][1], 'z': image_array[inds[0][0]][2]}
                template_points_with_neighbour.append(dict_template)
                image_points_with_neighbour.append(dict_image)
    
    template_matched_df = pd.DataFrame(template_points_with_neighbour, columns=['x', 'y', 'z'])
    image_matched_df = pd.DataFrame(image_points_with_neighbour, columns=['x', 'y', 'z'])
    colors = [plt.cm.Spectral(each)
              for each in np.linspace(0, 1, len(image_matched_df.z))]    
    random.shuffle(colors
        
    if plot:
        fig, ax = plot3Dstructure(image_matched_df, df_2=template_matched_df, colors=colors)
    return template_matched_df, image_matched_df
      
def rms_3D(df1, df2):
    '''calculates the RMSE between two sets of 3D data points'''
    return np.sqrt( np.sum((df1.z-df2.z)**2 +  (df1.x-df2.x)**2 +(df1.y-df2.y)**2  )/len(df1.z))

def kde1D(x, bandwidth=0.01):
    '''plots gaussian kernel density of 1D array'''   
    X=np.array(x)
    X = X.reshape([-1,1])
    size_x=np.max(x)-np.min(x)
    X_plot = np.linspace(np.min(x)-size_x*0.05, np.max(x)+size_x*0.05, 1000)[:, None]
    # Gaussian KDE
    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(X)
    
    # score_samples() returns the log-likelihood of the samples
    log_dens = kde.score_samples(X_plot)
    z = np.exp(log_dens)
    
    fig, ax = plt.subplots(1)
    ax.fill(X_plot[:, 0], z, fc='#AAAAFF')
    return fig,ax
    
def kde2D(x, y, bandwidth, xbins=100j, ybins=100j, **kwargs): 
    """Build 2D kernel density estimate (KDE)."""

    # create grid of sample locations (default: 100x100)
    xx, yy = np.mgrid[x.min():x.max():xbins, 
                      y.min():y.max():ybins]

    xy_sample = np.vstack([yy.ravel(), xx.ravel()]).T
    xy_train  = np.vstack([y, x]).T
    # Gaussian KDE
    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth, **kwargs)
    kde.fit(xy_train)

    # score_samples() returns the log-likelihood of the samples
    z = np.exp(kde.score_samples(xy_sample))
    return xx, yy, np.reshape(z, xx.shape)

def kde3D(x, y, z, bandwidth=0.01, xbins=100j, ybins=100j, zbins=100j, **kwargs): 
    """Build 2D kernel density estimate (KDE)."""

    # create grid of sample locations (default: 100x100)
    xx, yy, zz = np.mgrid[x.min():x.max():xbins, 
                      y.min():y.max():ybins,
                      z.min():z.max():zbins]

    xyz_sample = np.vstack([zz.ravel(), yy.ravel(), xx.ravel()]).T
    xyz_train  = np.vstack([z, y, x]).T
    # Gaussian KDE
    kde_skl = KernelDensity(kernel='gaussian', bandwidth=bandwidth, **kwargs)
    kde_skl.fit(xyz_train)
    
    # score_samples() returns the log-likelihood of the samples
    z = np.exp(kde_skl.score_samples(xyz_sample))
    return xx, yy, zz, np.reshape(z, xx.shape)

def investigate_neighbours_2D(atompos_df, plot_kde=True, max_dist=2, nhood_size=10, MIN_dist=0, n_neighbours=None): 
    ''' Returns array of nearest neighbours distances. '''
    
    atompos_array = atompos_df.as_matrix(columns = ['x_proj', 'z'])
    kdt = KDTree(atompos_array,  metric='euclidean')
    n_neighbours=n_neighbours if n_neighbours is not None else len(atompos_df.z)
    
    dist_arr=np.array([])
    for i in range(0, len(atompos_df.z)):
        dist, ind = kdt.query([atompos_array[i]], k=n_neighbours) 
        dist=dist[0]
        dist=dist[1:]
        dist=dist[dist<nhood_size]
        dist=dist[dist>MIN_dist]
        dist_arr = np.concatenate((dist_arr, dist))

    if plot_kde:
        fig,ax=kde1D(dist_arr, bandwidth=0.01)
    return layer_db_df, dist_arr

def investigate_neighbours_3D(atompos_df, check_sub_df=None, plot_hist=False, plot_kde=True,dbin = 1, max_dist=2, MIN_dist=0, nhood_size=10, n_neighbours=5): 
    ''' Plots histogram with all neighbour distances, and layer means 
    Returns layer_db_df with mean and std of each "layer" of neighbours 
    if check_sub_df is input, only these atoms are checked. '''
    
    atompos_array = atompos_df.as_matrix(columns = ['x', 'y', 'z'])
    kdt = KDTree(atompos_array,  metric='euclidean')
    n_neighbours=n_neighbours if n_neighbours is not None else len(atompos_df.z)
    
    dist_arr=np.array([])
    
    if check_sub_df is not None:
        check_array = check_sub_df.as_matrix(columns = ['x', 'y', 'z'])
        for i in range(0, len(check_sub_df.z)):
            dist, ind = kdt.query([check_array[i]], k=(n_neighbours+1)) 
            dist=dist[0]
            dist=dist[dist<nhood_size]
            dist=dist[dist>MIN_dist]
            dist_arr = np.concatenate((dist_arr, dist[1:]))
    else:
        for i in range(0, len(atompos_df.z)):
            dist, ind = kdt.query([atompos_array[i]], k=(n_neighbours+1)) 
            dist=dist[0]
            dist=dist[1:]
            dist=dist[dist<nhood_size]
            dist=dist[dist>MIN_dist]
            dist_arr = np.concatenate((dist_arr, dist))
            
    if plot_hist:
        plt.figure()
        plt.hist(dist_arr, bins=np.arange(np.min(dist_arr), np.max(dist_arr), dbin))

    ####bruke 1D DBSCAN paa dist_arr, returnere  array med means 
    X2d = np.zeros((len(dist_arr), 2))
    X2d[:,0] = dist_arr#.reshape(1, -1)
    
    db = DBSCAN(eps=max_dist, min_samples=1).fit(X2d)  
    
    labels=db.labels_
    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
   
    ##Reknar ut center som CM av kvar cluster..
    layer_list = []    
    for i in range(0, n_clusters_):
        layer = X2d[labels==i]
        no_atoms=len(layer[:,0])
        mean = np.mean(layer[:,0])
        std = np.std(layer[:,0])
        dict1 = {'layer_dist_mean': mean, 'layer_dist_std': std, 'no_atoms': no_atoms}
        layer_list.append(dict1)
        
    layer_db_df = pd.DataFrame(layer_list, columns=['layer_dist_mean', 'layer_dist_std', 'no_atoms'])
    
    #add to histogram
    if plot_hist:
        plt.scatter(layer_db_df.layer_dist_mean, np.zeros(len(layer_list)), s=20, color='r')
        
    if plot_kde:
        fig, ax=kde1D(dist_arr)
        if plot_hist:
            ax.scatter(layer_db_df.layer_dist_mean, np.zeros(len(layer_list)), s=20, color='r')
    
    return layer_db_df, dist_arr
